//----------------------------------------------------------------------------------------------------------------------
// Common shader model data (PLEASE KEEP THE VARIABLES IN CONSTANT BUFFERS AND STRUCTURES OF THIS FILE UNCHANGED)
//----------------------------------------------------------------------------------------------------------------------
#pragma pack_matrix(row_major) // Set the matrix packing alignment to row major
#define vLucent { 0, 0, 0, 0 } // Default fog color (initially translucent air)
#define MODULUS 2147483647U // Random generator modulus (hex number 0x7fffffff)

Texture2D s_pParticleTexture : register(t0); // Texture resource for particle effects (must use slot number 0)
SamplerState s_pLinearWrap : register(s0); // Linear filter and wrap address sampling (must use slot number 0)

cbuffer XCOVENANT : register(b0) // Constant buffer for every shader program (must use slot number 0)
{
   matrix s_mCameraViewProj;     // Transposed camera view matrix (3 rows) and projection vector (4th row)
   float3 s_vCameraPosition;     // Position of the current camera in world space
   float  s_fGameTime = 0;       // Current game time in seconds (for randomness)
   float4 s_vFogColor = vLucent; // Fog color (fog effect is controlled by alpha)
   float  s_fFogStart = 0.25;    // Fog start distance from camera in view space
   float  s_fFogRange = 1000;    // Fog coverage range from camera in view space
   float  s_fVariable = 0;       // Custom float variable defined by the application
   float  s_fReflection = 0;     // Mirror reflection ratio (zero means no reflection)
   float4 s_vReflection;         // Mirror reflection plane (quaternion is normalized)
};

cbuffer XPARTICLE : register(b1) // Constant buffer for the particle effects (must use slot number 1)
{
   float3 s_vPosition;           // Emitter particle position in world space
   float  s_fParticle = 0;       // Custom float variable defined by particle
   float3 s_vVelocity;           // Emitter particle velocity in world space
   uint   s_uParticle = 0;       // Custom uint variable defined by particle
   float3 s_vAcceleration;       // Global particle acceleration in world space
   uint   s_uRandomNumber = 1;   // Random uint number [1, 2147483646]
};

struct PARTICLE_VG_IN // Particle effects input data for vertex shader and geometry shader
{
   float3 vPosW : POSITION;      // Particle position in world space
   float  fTime : TIME;          // Particle time
   float3 vVelW : VELOCITY;      // Particle velocity in world space
   float  fSize : PSIZE;         // Particle size
   float3 vAccW : ACCELERATION;  // Particle acceleration in world space
   bool   bEmit : EMITTER;       // Particle is emitter
};

struct PARTICLE_PS_IN // Particle effects input data for pixel shader
{
   float4 vPosH : SV_Position;   // Particle position in homogeneous clip space
   float2 vTexC : TEXCOORD;      // Particle texture coordinates
   float  fInkA : COLOR;         // Particle color modulation
   float  fFogW : FOG;           // Particle fog weight
};

//----------------------------------------------------------------------------------------------------------------------
// Uniform random vector generator
//----------------------------------------------------------------------------------------------------------------------
float3 GetRandomVector(uint uSeed)
{
   float vRand[3];
   uint uRand = (s_uRandomNumber+uint(s_fGameTime)+uSeed)%(MODULUS-1)+1; // Range [1, 2147483646]
   for(uint i = 0; i < 3; i++)
   {
      uRand = 48271*(uRand%44488)-3399*(uRand/44488); // Run generator
      if(uRand > MODULUS) uRand += MODULUS; // Range [1, 2147483646]
      vRand[i] = 2*float(uRand-1)/(MODULUS-2)-1; // Range [-1, 1]
   }
   return float3(vRand);
}

//----------------------------------------------------------------------------------------------------------------------
// Transform vertex coordinates from world space to camera view space
//----------------------------------------------------------------------------------------------------------------------
float4 WorldToCameraView(float4 vPosW, float4 vPlane)
{
   float4 vPosV;
   if(s_fReflection) // Mirror reflection drawing mode is enabled
   {
      float fNorW = 2*dot(vPosW, vPlane);
      vPosW.xyz -= fNorW*vPlane.xyz;           // Camera view space coordinate system
   }                                           // Z                      Y
   vPosV.x = dot(vPosW, s_mCameraViewProj[0]); // |  Y (depth)           |  Z (depth)
   vPosV.y = dot(vPosW, s_mCameraViewProj[2]); // | /      =========>>>  | /
   vPosV.z = dot(vPosW, s_mCameraViewProj[1]); // |/       Swap y and z  |/
   vPosV.w = vPosW.w;                          // O------X               O------X
   return vPosV; // Coordinates y and z are now swapped after view transformation
}

//----------------------------------------------------------------------------------------------------------------------
// Transform vertex coordinates from camera view space to homogeneous clip space
//----------------------------------------------------------------------------------------------------------------------
float4 CameraViewToClip(float4 vPosV)
{
   float4 vPosH = float4(vPosV.xyz*s_mCameraViewProj[3].xyz, vPosV.w); // Assume orthographic projection
   if(s_mCameraViewProj[3].w) // Update components for perspective projection
   {
      vPosH.z += s_mCameraViewProj[3].w; // Depth in homogeneous clip space
      vPosH.w = vPosV.z; // Depth in view space for perspective divide
   }
   return vPosH;
}

//----------------------------------------------------------------------------------------------------------------------
// Compute current fog weight factor (not saturated)
//----------------------------------------------------------------------------------------------------------------------
float GetFogWeight(float3 vPosW)
{
   return (distance(vPosW, s_vCameraPosition)-s_fFogStart)/s_fFogRange;
}

//----------------------------------------------------------------------------------------------------------------------
// Compute current fogged color (fog weight is saturated)
//----------------------------------------------------------------------------------------------------------------------
float4 GetFoggedColor(float4 vColor, float fFogW)
{
   if(s_vFogColor.a) // Fog effect is enabled
   {
      fFogW = pow(saturate(fFogW), 1-s_vFogColor.a); // Fog effect increases with alpha
      vColor.xyz = lerp(vColor.xyz, s_vFogColor.xyz, fFogW);
   }
   return vColor;
}

//----------------------------------------------------------------------------------------------------------------------
// Compute mirror reflected color (fog color is considered)
//----------------------------------------------------------------------------------------------------------------------
float4 GetMirroredColor(float4 vColor, float fFogW)
{
   vColor = GetFoggedColor(vColor, fFogW); // Consider possible fogging effect
   if(s_fReflection) vColor.a *= s_fReflection; // Modify alpha for mirror reflection drawing mode
   return vColor;
}
