//----------------------------------------------------------------------------------------------------------------------
// Guideline:                                                                                                         //
// The particle effect interface is subject to change                                                                 //
// Each effect must define a stream shader followed by render shaders                                                 //
// Reference:                                                                                                         //
// Frank D. Luna (2008), Introduction to 3D Game Programming with Direct3D 10.0                                       //
// http://www.amazon.com/Introduction-3D-Game-Programming-DirectX/dp/1598220535                                       //
// http://www.d3dcoder.net/d3d10.aspx                                                                                 //
//----------------------------------------------------------------------------------------------------------------------
#include "axshared.hlsl" // This shared file must be included by all particle effects

//----------------------------------------------------------------------------------------------------------------------
// The stream-out geometry shader is just responsible for emitting new particles and destroying old particles
//----------------------------------------------------------------------------------------------------------------------
[maxvertexcount(2)]
void GS_Stream(point PARTICLE_VG_IN gIn[1], inout PointStream<PARTICLE_VG_IN> gStream)
{
   gIn[0].vPosW += (gIn[0].vVelW+0.5*gIn[0].vAccW*gIn[0].fTime)*gIn[0].fTime;
   gIn[0].fTime += 0.01; // Use a different time scale for particle life
   if(gIn[0].bEmit)
   {
      if(gIn[0].fTime > 0.005) // Time to emit a new particle
      {
         PARTICLE_VG_IN p;
         p.vPosW = s_vPosition+0.2*GetRandomVector(5);
         p.fTime = 0;
         p.vVelW = 1.5*s_vVelocity*GetRandomVector(3);
         p.fSize = 3;
         p.vAccW = s_vAcceleration*(1+float3(4.5, 4.5, 1.5)*GetRandomVector(1));
         p.bEmit = false;
         gStream.Append(p);
         gIn[0].fTime = 0; // Reset the time to emit
      }
      gStream.Append(gIn[0]); // Keep particle emitter
   }
   else if(gIn[0].fTime < 0.15) // Specify conditions to keep particle
   {
      gStream.Append(gIn[0]); // Keep emitted particle
   }
}

//----------------------------------------------------------------------------------------------------------------------
// Geometry shader for particle rendering
//----------------------------------------------------------------------------------------------------------------------
[maxvertexcount(4)]
void GS_Render(point PARTICLE_VG_IN gIn[1], inout TriangleStream<PARTICLE_PS_IN> gStream)
{
   static const float2 vQuad[4] = { { 0, 0 }, { 0, 1 }, { 1, 0 }, { 1, 1 } };
   if(!gIn[0].bEmit) // Do not render emitter particles
   {
      // Compute a world matrix so that billboard faces the camera
      matrix W;
      float3 look  = normalize(gIn[0].vPosW-s_vCameraPosition);
      float3 right = cross(look, float3(0, 0, 1));
      float3 up    = cross(right, look);
      W[0] = float4(right,        0);
      W[1] = float4(look,         0);
      W[2] = float4(up,           0);
      W[3] = float4(gIn[0].vPosW, 1);

      // Compute four triangle strip vertices (quad) in local space
      float4 v[4];
      float fHalfWidth = 0.5*gIn[0].fSize;
      v[0] = float4(-fHalfWidth, 0, +fHalfWidth, 1);
      v[1] = float4(-fHalfWidth, 0, -fHalfWidth, 1);
      v[2] = float4(+fHalfWidth, 0, +fHalfWidth, 1);
      v[3] = float4(+fHalfWidth, 0, -fHalfWidth, 1);

      // Transform vertices to homogeneous clip space and output a triangle strip
      PARTICLE_PS_IN gOut;
      float fAlpha = 1-smoothstep(0, 1, gIn[0].fTime); // Fade particle color with time
      for(uint i = 0; i < 4; i++)
      {
         gOut.vPosH = CameraViewToClip(WorldToCameraView(mul(v[i], W), s_vReflection));
         gOut.vTexC = vQuad[i];
         gOut.fInkA = fAlpha;
         gOut.fFogW = GetFogWeight(gIn[0].vPosW);
         gStream.Append(gOut);
      }
   }
}

//----------------------------------------------------------------------------------------------------------------------
// Pixel shader for particle rendering
//----------------------------------------------------------------------------------------------------------------------
float4 PS_Render(PARTICLE_PS_IN pIn) : SV_Target
{
   float4 vColor = s_pParticleTexture.Sample(s_pLinearWrap, pIn.vTexC)*pIn.fInkA;
   return saturate(GetMirroredColor(vColor, pIn.fFogW)); // Return saturated pixel color
}
