//----------------------------------------------------------------------------------------------------------------------
// Guideline:                                                                                                         //
// The particle effect interface is subject to change                                                                 //
// Each effect must define a stream shader followed by render shaders                                                 //
// Reference:                                                                                                         //
// Frank D. Luna (2008), Introduction to 3D Game Programming with Direct3D 10.0                                       //
// http://www.amazon.com/Introduction-3D-Game-Programming-DirectX/dp/1598220535                                       //
// http://www.d3dcoder.net/d3d10.aspx                                                                                 //
//----------------------------------------------------------------------------------------------------------------------
#include "axshared.hlsl" // This shared file must be included by all particle effects

//----------------------------------------------------------------------------------------------------------------------
// The stream-out geometry shader is just responsible for emitting new particles and destroying old particles
//----------------------------------------------------------------------------------------------------------------------
[maxvertexcount(6)]
void GS_Stream(point PARTICLE_VG_IN gIn[1], inout PointStream<PARTICLE_VG_IN> gStream)
{
   gIn[0].vPosW += (gIn[0].vVelW+0.5*gIn[0].vAccW*gIn[0].fTime)*gIn[0].fTime;
   gIn[0].fTime += 0.01; // Use a different time scale for particle life
   if(gIn[0].bEmit)
   {
      if(gIn[0].fTime > 0.005) // Time to emit a new particle
      {
         for(uint i = 0; i < 5; i++)
         {
            PARTICLE_VG_IN p;
            p.vPosW = s_vPosition+35*GetRandomVector(100*i);
            p.fTime = 0;
            p.vVelW = float3(0, 0, 0);
            p.fSize = 1;
            p.vAccW = s_vAcceleration;
            p.bEmit = false;
            gStream.Append(p);
         }
         gIn[0].fTime = 0; // Reset the time to emit
      }
      gStream.Append(gIn[0]); // Keep particle emitter
   }
   else if(gIn[0].fTime < 7) // Specify conditions to keep particle
   {
      gStream.Append(gIn[0]); // Keep emitted particle
   }
}

//----------------------------------------------------------------------------------------------------------------------
// Geometry shader for particle rendering
//----------------------------------------------------------------------------------------------------------------------
[maxvertexcount(2)]
void GS_Render(point PARTICLE_VG_IN gIn[1], inout LineStream<PARTICLE_PS_IN> gStream)
{
   if(!gIn[0].bEmit) // Do not render emitter particles
   {
      // Slant line in acceleration direction and add some random variation
      float3 p0 = gIn[0].vPosW;
      float3 p1 = gIn[0].vPosW+gIn[0].fSize*s_vAcceleration+float3(1.5, 0.1, 0)*cos(0.5*s_fGameTime);

      // Transform vertices to homogeneous clip space and expand the point into a line
      PARTICLE_PS_IN v[2];
      v[0].vPosH = CameraViewToClip(WorldToCameraView(float4(p0, 1), s_vReflection));
      v[0].vTexC = float2(0, 0);
      v[0].fInkA = 0.9375;
      v[0].fFogW = GetFogWeight(p0);
      gStream.Append(v[0]);
      v[1].vPosH = CameraViewToClip(WorldToCameraView(float4(p1, 1), s_vReflection));
      v[1].vTexC = float2(1, 1);
      v[1].fInkA = 0.8125;
      v[1].fFogW = GetFogWeight(p1);
      gStream.Append(v[1]);
   }
}

//----------------------------------------------------------------------------------------------------------------------
// Pixel shader for particle rendering
//----------------------------------------------------------------------------------------------------------------------
float4 PS_Render(PARTICLE_PS_IN pIn) : SV_Target
{
   float4 vColor = s_pParticleTexture.Sample(s_pLinearWrap, pIn.vTexC);
   return saturate(GetMirroredColor(vColor, pIn.fFogW)); // Return saturated pixel color
}
